/*F File_name: 
/*F	hosking.c
/*
/*P Purpose:
/*P 	The purpose of this program is to implement Hosking's frational ARIMA
/*P 	model to generate synthetic long-range dependent time series.
/*
/*U Usage:
/*U     cc -o hosking hosking.c -lm -ldl
/*U     hosking 0.0 > H5.dat   (for Hurst value of 0.5)
/*U     hosking 0.1 > H6.dat   (for Hurst value of 0.6)
/*U     hosking 0.2 > H7.dat   (for Hurst value of 0.7)
/*U     hosking 0.3 > H8.dat   (for Hurst value of 0.8)
/*U     hosking 0.4 > H9.dat   (for Hurst value of 0.9)
/*
/*A Author:
/*A  	Zhiang Deng  (and modified by Carey Williamson, July 1996)
/*A                  (and modified by Ram Balakrishnan, February 1999)
/*
/*D Date:
/*D 	July 1995
/*
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define LENGTH  10000 /* Length of trace */

/* Carey modified this to generate a zero-mean sequence */
/* with values in the range [-1.0,1.0]. We can always   */
/* take and transform these values to other range later */

#define MIN_IN  -1   /* Minimum of output */
#define MAX_IN  1    /* Maximum of output */

#define PI      3.14159265
#ifndef RAND_MAX
#define RAND_MAX	0x7fffffff
#endif

/* #define VERBOSE 1 */

float rho[LENGTH], phi0[LENGTH], phi1[LENGTH], x[LENGTH];

float normal();

main(argc, argv)
int argc;
char **argv;
{
  float d;			/* H = d + 0.5 */
  long i,j;
  float m0 = 0.0, v0 = 1.0;
  float n0, n1, d0, d1, m1, v1;
  float xmax = -1.0e9, xmin = 1.0e9, mean = 0.0;
  long range;
  int clength;

  if(argc < 3)
  {
    printf("Usage: hosking d_value length > output_file\n");
    printf("  where d_value is in the range [0.0,0.5], and H = d + 0.5\n");
    exit(-1);
  }

  d = atof(argv[1]);
  if( (d < 0.0) || (d > 0.5) )
    {
      printf("The value for d must be in the range [0.0,0.5]\n");
      exit(-1);
    }

  clength = atoi(argv[2]);
  if( (clength < 0) || (clength > 10000) )
    {
      printf("No of values should be in the range [0,10000]\n");
      exit(-1);
    }

  srand(time(0));

  rho[0] = 1.0;
  for (i = 1; i < clength; i++)
      rho[i] = rho[i-1]*(i-1+d)/(i-d);

  /* Carey added this initialization for phi0  May 8/98 */
  for (i = 0; i <= clength; i++)
    phi0[i] = 0.0;

  x[0] = normal(m0, v0);
  if (x[0] < xmin) xmin = x[0];
  if (x[0] > xmax) xmax = x[0];
  mean += x[0];
  n0 = 0.0;
  d0 = 1.0;

  for (i = 1; i < clength; i++)
  {
     n1 = rho[i];
     for (j = 1; j < i; j++)
	n1 -= phi0[j]*rho[i-j];

     /* If d1 cause floating-point exception, underflow. */ 
     if( fabs(d0 - n0) < 1.0e-8) 
     {
       puts("d1 is underflow!");
       d1 = 1.0e-8;
     } 
     else
       d1 = d0 - n0*n0/d0;

     phi1[i] = n1/d1;

     for (j = 1; j < i; j++)
	phi1[j] = phi0[j] - phi1[i]*phi0[i-j];

     m1 = 0.0;
     for (j = 1; j <= i; j++)
	m1 += phi1[j]*x[i-j];

     v1 = (1.0 - phi1[i]*phi1[i])*v0;

     x[i] = normal(m1, v1);

     if (x[i] < xmin) xmin = x[i];
     if (x[i] > xmax) xmax = x[i];
     mean += x[i];

     v0 = v1;
     n0 = n1;
     d0 = d1;
     for (j = 1; j <= i; j++)
	phi0[j] = phi1[j];

  }

  mean /= (float)clength;

#ifdef VERBOSE
  printf("Range: %f to %f\t Mean= %f\n", xmin, xmax, mean);
#endif VERBOSE

  range = MAX_IN - MIN_IN;
  xmax = xmax - xmin;
  for(i=0; i<clength; i++)
  {
    printf("%f\n", (float)(1.0 * range * (x[i] - xmin) / xmax + MIN_IN) );
  }

  return(0);
}

float normal(m, v)
float m,v;
{
  float r1 = (float)rand()/RAND_MAX;
  float r2 = (float)rand()/RAND_MAX;
  float z =  sqrt(-2.0*log(r1))*cos(2.0*PI*r2);
  return m+sqrt(v)*z;
}
